// ENTITY_3DS Class    (c) by Heinrich Tillack 2002
// v 0 (21.10.02): initial release

#include "entity3ds.h"

#include <vector>
#include <GL/gl.h>
extern "C" {
	#include "all.h"
}

ENTITY_3DS::ENTITY_3DS(){

    model=NULL;
}

ENTITY_3DS::~ENTITY_3DS()
{
    if(model!=NULL){
        printf("\nENTITY_3DS: %s deleted\n",name);
        delete name;
        for(int i = 0; i < model->numOfObjects; i++)
        {
            // Free the faces, normals, vertices, and texture coordinates.
            delete [] model->pObject[i].pFaces;
            #ifdef USE_NORMALS
            delete [] model->pObject[i].pNormals;
            #endif
            delete [] model->pObject[i].pVerts;
            delete [] model->pObject[i].pTexVerts;

        }

    }
}

GLvoid ENTITY_3DS::FreeTextures(){
    if(model==NULL) return;
    for(int i = 0; i < model->numOfMaterials; i++)
        delete_tex(model->pMaterials[i].materialId) ;

}




GLvoid ENTITY_3DS::SetModel(t3DModel *tmodel,char *modelname){

    name =new char[64];
    memcpy(name, modelname, 64);

    model=tmodel;

    #if 1
    Vec3 _center;

    //compute  center
    for(int i=0;i<model->numOfObjects;i++){
        struct t3DObject *pObject=&model->pObject[i];
        _center.Reset();
        for(int j=0;j<pObject->numOfVerts;j++){
            _center+=pObject->pVerts[j];
        }
        _center/=pObject->numOfVerts;

        pObject[i].sphere.SetCenter(_center);


    }

    //compute  BoundingSphereRadius

    for(int i=0;i<model->numOfObjects;i++){
        struct t3DObject *pObject=&model->pObject[i];
        _center=pObject->sphere.GetCenter();
        float radius=0.0f,r2; float x,y,z;
        for(int j=0;j<pObject->numOfVerts;j++){
            x=pObject->pVerts[j].x-_center.x;
            y=pObject->pVerts[j].y-_center.y;
            z=pObject->pVerts[j].z-_center.z;

            r2 = (x * x + y * y + z * z);
            if(r2 > radius)
                radius = r2;
        }

        pObject[i].sphere.SetRadius( fsqrt(radius) );

        printf("3DS Object Nr %d, Center %f,%f,%f \n", i,pObject[i].sphere.center.x, pObject[i].sphere.center.y, pObject[i].sphere.center.z);
        printf("3DS Object Nr %d, Radius %f \n",  i,pObject[i].sphere.GetRadius());

    }
#endif
}

//FIXME: Currently only textured objects are allowed
GLvoid ENTITY_3DS::Draw()
{


    Context		&context = Context::Get();



    Apply();


    #define g_3Dmodel model

    Vec3 pos;
    /* Expect CCW verts */
    glFrontFace(GL_CCW);

    // Since we know how many objects our model has, go through each of them.
    for(int i = 0; i < g_3Dmodel->numOfObjects; i++)
    {
        // Make sure we have valid objects just in case. (size() is in the vector class)
        if(g_3Dmodel->pObject.size() <= 0) break;

        // Get the current object that we are displaying
        t3DObject *pObject = &g_3Dmodel->pObject[i];

        pos=Position-pObject->sphere.GetCenter();
        //pos=pObject->Center;

        int ClipState=context.GetFrustrum()->SphereInFrustum(pos.x,pos.y,pos.z, pObject->sphere.GetRadius());

        if(ClipState==FULL_OUTSIDE) {
            //   printf("3DS Object %d is outside\n",i);
            continue;
        }
        else if(ClipState==FULL_INSIDE)
            glDisable(GL_KOS_NEARZ_CLIPPING);
        else
            glEnable(GL_KOS_NEARZ_CLIPPING);

        //         // Check to see if this object has a texture map, if so bind the texture to it.
        //   if(pObject->bHasTexture) {

        //             //             // Turn off texture mapping and turn on color
        //             glEnable(GL_TEXTURE_2D);


        //             //             // 			// Reset the color to normal again
        //             glColor3f(1, 1, 1);

        //             // Bind the texture map to the object by it's materialID

        //  switch_tex(g_3Dmodel->g_Texture[pObject->materialID]);
        switch_tex(pObject->materialID);

        //  }
        //else {

        //             // Turn off texture mapping and turn on color
        //             glDisable(GL_TEXTURE_2D);

        //             // Reset the color to normal again
        //             glColor3f(1, 1, 1);
        //         }

        // This determines if we are in wireframe or normal mode
        glBegin(GL_TRIANGLES);

        // Go through all of the faces (polygons) of the object and draw them
        for(int j = 0; j < pObject->numOfFaces; j++)
        {
            // Go through each corner of the triangle and draw it.
            for(int whichVertex = 0; whichVertex < 3; whichVertex++)
            {
                // Get the index for each point of the face
                int index = pObject->pFaces[j].vertIndex[whichVertex];

                // Give OpenGL the normal for this vertex.
                //FIXME
                //glNormal3f(pObject->pNormals[ index ].x, pObject->pNormals[ index ].y, pObject->pNormals[ index ].z);

                //                 // If the object has a texture associated with it, give it a texture coordinate.
                //                 if(pObject->bHasTexture) {

                //                     // Make sure there was a UVW map applied to the object or else it won't have tex coords.
                //                     if(pObject->pTexVerts) {


                kglTexCoord(pObject->pTexVerts[ index ].x, pObject->pTexVerts[ index ].y);
                //                     }
                //                 } else {

                //                     // Make sure there is a valid material/color assigned to this object.
                //                     // You should always at least assign a material color to an object,
                //                     // but just in case we want to check the size of the material list.
                //                     // if the size is at least one, and the material ID != -1,
                //                     // then we have a valid material.
                //                     if(g_3Dmodel->pMaterials.size() && pObject->materialID >= 0)
                //                     {
                //                         // Get and set the color that the object is, since it must not have a texture
                //                         char *pColor = g_3Dmodel->pMaterials[pObject->materialID].color;

                //                         // Assign the current color to this model

                //                         glColor3ub(pColor[0], pColor[1], pColor[2]);
                //                     }
                //                 }

                // Pass in the current vertex of the object (Corner of current face)
                kglVertex3f(pObject->pVerts[ index ].x, pObject->pVerts[ index ].y, pObject->pVerts[ index ].z);
            }
        }

        glEnd();								// End the drawing
    }


}
