// ENTITY_IMD Class
// (c) by Heinrich Tillack 2002 (http://a128.x15.org)
// under GPL or new BSD license
//
// with IMD code taken from Iris3D V0.7
// Author : Olivier Chatry - Epitech Console Laboratory
//           (http://www.epitech.net/labconsole/)
//
// v 0 (21.10.02): initial release
// v 0.1		 : BoundingBox code added and updted BoundingSphere

#include "entityImd.hpp"
#include "Vec3.h"

#include "model3d.h"

#include <kos.h>
#include <gl/gl.h>

extern "C" {
	#include "all.h"
}



ENTITY_IMD::ENTITY_IMD(){
    model.numOfMaterials=0;
    model.numOfObjects=0;
    model.numOfIndicies=0;
    //model.Reset();

}

ENTITY_IMD::~ENTITY_IMD(){


    for(int i = 0; i < model.numOfMaterials; i++)
        delete_tex(model.pMaterials[i].textureId) ;

    printf("\nENTITY_IMD: %s texture deleted\n",_name);
#if 1

    //Not working ?!
    for(int i = 0; i < model.numOfObjects; i++)
    {

        t3DObject *pObject = &model.pObject[i];
        // Free the strips, normals, vertices, and texture coordinates.
        for(int j = 0; j < pObject->numOfSections; j++)
        {

            delete pObject->pStrip[j]._indice;

        }
        printf("\nENTITY_IMD:free1\n");
            #ifdef USE_NORMALS
        delete [] &model.pObject[i].pNormals;
            #endif
        delete [] &model.pObject[i].pVerts;
        printf("\nENTITY_IMD:free2\n");
        delete [] &model.pObject[i].pTexVerts;
        printf("ENTITY_IMD: %s object Nr %d deleted\n",_name,i);
    }



    printf("ENTITY_IMD: %s deleted\n",_name);
    delete _name;

    //model.Free();

    #endif
}

void	ENTITY_IMD::LoadMaterial(imd_object_t *imd_object, char *file_path)
{
    for (int i = 0; i < imd_object->imd_object_header.num_material; i++)
    {

        tMaterialInfo newTexture = {0};
        // Increase the number of materials
        model.numOfMaterials++;

        char	*file_path_and_name = GetFilePath(file_path, imd_object->imd_material[i].file_name);

        newTexture.textureId=auto_load_texture(file_path_and_name);

        strcpy( newTexture.strFile,file_path_and_name);

        delete [] file_path_and_name;

        newTexture.materialId=imd_object->imd_material[i].material_id;

        model.pMaterials.push_back(newTexture);
    }
}



void	ENTITY_IMD::LoadMesh(imd_object_t *imd_object,float scale=1)
{

    model.numOfObjects=imd_object->imd_object_header.num_mesh;

    printf("\n\nENTITY_IMD: %s has %d mesh(s)\n",_name,model.numOfObjects);

    for (int i = 0; i < model.numOfObjects; i++){

        struct t3DObject newMesh ;

        //   printf("\nloading IMD Mesh %d...",i);

        newMesh.numOfFaces=0; //cause we have strips no TRIANGLES

        imd_mesh_t &imd_mesh=imd_object->imd_mesh[i];

        int count = imd_mesh.imd_mesh_header.num_vertex * _num_anim;
        int _num_vertex = imd_mesh.imd_mesh_header.num_vertex;

        newMesh.numOfVerts=_num_vertex;

        //    printf("vert=%d...",_num_vertex);

        //load verts
	#ifdef USE_NORMALS
        uint32 *color = new uint32 [count];
        Vec3 *normal			= new Vec3[count];
	#endif
        Vec3 *vertex			= new Vec3[count];

        // 	_trans_normal	= new Vec3[_num_vertex];
        // 	_trans_vertex	= new Vec3[_num_vertex];
        //_2d_vertex		= new vertex2dtl[_num_vertex];
        //iris_object->num_vertex += imd_mesh.imd_mesh_header.num_vertex;

        Vec3 _bmin;
        Vec3 _bmax;

        _bmin.set(999999.0f,999999.0f,999999.0f);
        _bmax.set(-999999.0f,-999999.0f,-999999.0f);


        for (int i = 0; i < count; i++)
        {
            vertex[i].x = scale*imd_mesh.imd_vertex[i].pos[0];
            vertex[i].y = scale*imd_mesh.imd_vertex[i].pos[1];
            vertex[i].z = scale*imd_mesh.imd_vertex[i].pos[2];

		#ifdef USE_NORMALS

            normal[i].x = imd_mesh.imd_vertex[i].normal[0];
            normal[i].y = imd_mesh.imd_vertex[i].normal[1];
            normal[i].z = imd_mesh.imd_vertex[i].normal[2];
            //_normal[i].w = 1.0f;
            normal[i].Normalize();

            color[i] = imd_mesh.imd_vertex[i].color;
		#endif

            //update of AABB Box infos

            if (vertex[i].x < _bmin.x)
                _bmin.x = vertex[i].x;
            else if (vertex[i].x > _bmax.x)
                _bmax.x = vertex[i].x;

            if (vertex[i].y < _bmin.y)
                _bmin.y = vertex[i].y;
            else if (vertex[i].y > _bmax.y)
                _bmax.y = vertex[i].y;

            if (vertex[i].z < _bmin.z)
                _bmin.z = vertex[i].z;
            else if (vertex[i].z > _bmax.z)
                _bmax.z = vertex[i].z;


        }

        //Set the AABB info
        newMesh.box.SetMinMax(_bmin,_bmax);

        //Center of the Sphere constructed from the AABB!
        newMesh.sphere.SetCenter(newMesh.box.GetCenter());
        newMesh.sphere.SetRadius(newMesh.box.GetRadius());


        newMesh.pVerts=vertex;


	#ifdef USE_NORMALS
        newMesh.pNormals=normal;
        newMesh.color=color;

	#endif


        //load UV coords

        CVector2  *TexVerts=new CVector2[count];

        for (int i = 0; i < _num_vertex; i++)
        {
            TexVerts[i].x = imd_mesh.imd_vertex[i].uv[0];
            TexVerts[i].y = imd_mesh.imd_vertex[i].uv[1];
            //_2d_vertex[i].argb = imd_mesh.imd_vertex[i].color;
            //_2d_vertex[i].oargb = imd_mesh.imd_vertex[i].color;
        }

        newMesh.pTexVerts=TexVerts;
        ///////

        newMesh.numOfSections=imd_mesh.imd_face.num_section;

        IrisSection *_section = new IrisSection[newMesh.numOfSections];

        memcpy(newMesh.strName, imd_mesh.imd_mesh_header.name, IMD_MAX_NAME);

        for (int i = 0; i < imd_object->imd_object_header.num_material; i++)
            if ( model.pMaterials[i].materialId == imd_mesh.imd_mesh_header.material_id)
                newMesh.materialID = model.pMaterials[i].textureId;

        newMesh.bHasTexture=1;
        /////////////////

        //   printf("strips=%d...",newMesh.numOfSections);

        int i;

        if (newMesh.numOfSections > 0)
        {
            int	num_indice;
            unsigned short *indices;
            imd_face_section_s *src_section;

            IrisSection *dst_section;

            src_section = &imd_mesh.imd_face.imd_section[0];
            num_indice = 0;
            for (i = 0; i < newMesh.numOfSections; i++) // Total size should be precalculated in IMD
            {
                num_indice += src_section->num_indice;
                src_section++;
            }

            model.numOfIndicies = num_indice;

            indices = new unsigned short[num_indice];

            //   printf("indicies=%d...",num_indice);


            src_section = &imd_mesh.imd_face.imd_section[0];
            dst_section = &_section[0];

            newMesh.pStrip=dst_section;


            num_indice = 0;
            for (i = 0; i < newMesh.numOfSections; i++)
            {
                dst_section->_indice = indices + num_indice;
                dst_section->_num_indice = src_section->num_indice;
                memcpy(dst_section->_indice, src_section->indice, dst_section->_num_indice * sizeof (unsigned short));
                num_indice += src_section->num_indice;
                src_section++;
                dst_section++;
            }
        }
        /////
        model.pObject.push_back(newMesh);
    }
}


//Load and Convert IMD model to t3dmodel !
//Load also the PNG textures
GLvoid  ENTITY_IMD::Load( char *file_name,char *file_path,float scale=1)
{
    imd_object_t	*imd_object;

    _file_name = GetFilePath(file_path, file_name);
    imd_object = load_imd(_file_name);

    if (imd_object == NULL) {
        assert(0);
        return ;
    }

    delete _file_name;

    _name = new char[IMD_MAX_NAME];
    memcpy(_name, imd_object->imd_object_header.name, IMD_MAX_NAME);
    _num_anim = imd_object->imd_object_header.num_anim;
    LoadMaterial(imd_object, file_path);
    LoadMesh(imd_object,scale);
    //     LoadTag(imd_object);
    free_imd(imd_object);

    for(int i=0;i<model.numOfObjects;i++){
        struct t3DObject *pObject=&model.pObject[i];

        //         printf("\nIMD Object Nr %d, Center %f,%f,%f \n", i,pObject[i].sphere.center.x, pObject[i].sphere.center.y, pObject[i].sphere.center.z);
        //          printf("IMD Object Nr %d, Radius %f \n",  i,pObject[i].sphere.GetRadius());

    }

}


GLvoid  ENTITY_IMD::Draw(){


    Context		&context = Context::Get();



    glTranslatef(Position.x,Position.y,Position.z);

    //HACK: to fix the coords to KGL coords x,y,z ;y=up


	glRotatef( -90.0f, 1.0f, 0.0f, 0.0f );
	glRotatef( -90.0f, 0.0f, 0.0f, 1.0f );


 	glRotatef( Delta_y, 0.0f, 0.0f, 1.0f );
 	glRotatef( Delta_z, 0.0f, 1.0f, 0.0f );
	glRotatef( Delta_x, 1.0f, 0.0f, 0.0f );


//    Apply();


    #define g_3Dmodel model

    Vec3 pos;
    /* Expect CW verts */
    glFrontFace(GL_CW);

    // Since we know how many objects our model has, go through each of them.
    for(int i = 0; i < g_3Dmodel.numOfObjects; i++)
    {

        // Get the current object that we are displaying
        t3DObject *pObject = &g_3Dmodel.pObject[i];

        /////FRUSTUM CULLING
        pos=Position-pObject->sphere.GetCenter();
        //pos=pObject->Center;

        int ClipState=context.GetFrustrum()->SphereInFrustum(pos.x,pos.y,pos.z, pObject->sphere.GetRadius());

        if(ClipState==FULL_OUTSIDE) {
            //   printf("IMD Object %d is outside\n",i);
            continue;
        }
        else if(ClipState==FULL_INSIDE)
            glDisable(GL_KOS_NEARZ_CLIPPING);
        else
            glEnable(GL_KOS_NEARZ_CLIPPING);

        ///////

        //  printf("IMD Draw Object Nr %d...",i);
        uint32 vertex_index = GetCurrentAnim() * pObject->numOfVerts;

        uint32 vertex_index2 = GetNextAnim() * pObject->numOfVerts;

        float blendcount=(float)GetFPS();
        float blendpos=(float)GetFrameCount();

        switch_tex(pObject->materialID);


        Vec3 *verts;
        Vec3 *verts2;
        CVector2 *uv;

        //for each Strip section

        for(int j = 0; j < pObject->numOfSections; j++)
        {
            // Get the index for each point of the face
            int count = pObject->pStrip[j]._num_indice;
            uint16 *index=pObject->pStrip[j]._indice;

            // Get the index for each point of the face

            verts=pObject->pVerts+vertex_index;
            verts2=pObject->pVerts+vertex_index2;
            uv=pObject->pTexVerts;

            glBegin (GL_TRIANGLE_STRIP);

            for(int a=0;a<count;a++){

                uint16 i=*index++;

                float    x1=verts[i].x;
                float x=x1-((x1-verts2[i].x)/blendcount)*blendpos;


                float    y1=verts[i].y;
                float y=y1-((y1-verts2[i].y)/blendcount)*blendpos;

                float    z1=verts[i].z;
                float z=z1-((z1-verts2[i].z)/blendcount)*blendpos;

                kglTexCoord(uv[ i ].x, uv[ i ].y);

                kglVertex3f(x,y,z);


            }

            glEnd();

        }

    }


}
