// (c) by Heinrich Tillack 2002 (http://a128.x15.org)
// based on code my Kamka and md2Viewer by Chalumbasoft?!
// under GPL or new BSD license
// v 0 (21.10.02): initial release

#include <kos.h>
#include <GL/gl.h>
#include <dc/fmath.h>

#define fabs(x)	((x)<0?-(x):(x))

#include "all.h"


typedef float vec_t;
typedef vec_t vec3_t[3];
typedef vec_t vec2_t[2];

#ifdef MD2_LIGHT

#define NUMVERTEXNORMALS 162

float avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
        };


// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anormtab.h"
    ;

float	*shadedots = r_avertexnormal_dots[0];

vec3_t	shadevector;
float	shadelight[3];

#endif


struct md2_model_node *model_list=NULL;



int find_frame(struct md2_model_node *model, char *name) {
    int i;
    //DBG("Searching for: %s...",name);
    for(i=0;i<model->framecount;i++) {
        //DBG("Compare: \"%s\" and \"%s\"\n",name,model->frames[i].name);
        if(!strcmp(name,model->frames[i].name)) {
            //DBG("match!\n");
            return i;
        }
    }
    assert(0);
    return -1;
}

struct md2_model_node *get_model(char *filename) {
    //     struct md2_model_node *current_model=model_list;
    //     while(current_model!=NULL) {
    //         if(!strcmp(current_model->filename,filename)) {
    //             return current_model;
    //         }
    //         current_model=current_model->next;
    //     }
    return NULL;
}

struct md2_model_node *load_md2(char *filename,float scale) {
    int fd;
    int i,j;
    short s,t;
    struct md2_model_hdr model_hdr;
    struct md2_frame_hdr frame_hdr;
    struct md2_pt point;
    struct md2_model_node *current_model=NULL;


    fd=fs_open(filename,O_RDONLY);




    current_model=(struct md2_model_node *)malloc(sizeof(struct md2_model_node));

    assert(current_model!=NULL);

    current_model->filename=(char *)malloc(strlen(filename));

    strcpy(current_model->filename,filename);

    fs_read(fd,&model_hdr,sizeof(struct md2_model_hdr));

    current_model->framecount=model_hdr.numFrames;
    current_model->trianglecount=model_hdr.numTriangles;
    current_model->verticount=model_hdr.numVertices;

    printf("MD2 has %d skins, with width=%d and height=%d\n",
           model_hdr.numSkins,model_hdr.skinWidth,model_hdr.skinHeight);

    printf("Loading %i frames of %i vertices...\n",model_hdr.numFrames,model_hdr.numVertices);

    current_model->frames=(struct frame *)malloc(sizeof(struct frame) * model_hdr.numFrames);
    for(i=0;i<model_hdr.numFrames;i++) {
        current_model->frames[i].points=(struct pt *)malloc(sizeof(struct pt) * model_hdr.numVertices);
        fs_seek(fd,model_hdr.offsetFrames+(model_hdr.frameSize*i),SEEK_SET);

        fs_read(fd,&frame_hdr,sizeof(struct md2_frame_hdr));
        strcpy(current_model->frames[i].name,frame_hdr.name);
        //  printf("Frame name: %s\n",frame_hdr.name);

        for(j=0;j<model_hdr.numVertices;j++) {
            fs_read(fd,&point,sizeof(struct md2_pt));

            current_model->frames[i].points[j].x=scale*(point.vertex[0] * frame_hdr.scale[0] + frame_hdr.translate[0]);
            current_model->frames[i].points[j].y=scale*(point.vertex[2] * frame_hdr.scale[2] + frame_hdr.translate[2]);
            current_model->frames[i].points[j].z=-scale*(point.vertex[1] * frame_hdr.scale[1] + frame_hdr.translate[1]);

#ifdef MD2_LIGHT
            float nx = avertexnormals[point.lightNormalIndex][0];
            float ny = avertexnormals[point.lightNormalIndex][2];
            float nz = -avertexnormals[point.lightNormalIndex][1];


            float l = fsqrt(nx*nx+ny*ny+nz*nz);

            if (l != 0)
            {
                nx /= l;
                ny /= l;
                nz /= l;
            }


            current_model->frames[i].points[j].nx=nx;
            current_model->frames[i].points[j].ny=ny;
            current_model->frames[i].points[j].nz=nz;

            // normals and vertexes come from the frame list
            current_model->frames[i].points[j].l = shadedots[point.lightNormalIndex];

#endif
        }
    }

    //load up U and V values
    fs_seek(fd,model_hdr.offsetTexCoords,SEEK_SET);
    printf("Loading %i texture coordinates...\n",model_hdr.numTexCoords);
    current_model->texcoord=(struct tex_coord *)malloc(sizeof(struct tex_coord)*model_hdr.numTexCoords);
    for(i=0;i<model_hdr.numTexCoords;i++) {
        fs_read(fd,&s,sizeof(short));
        fs_read(fd,&t,sizeof(short));
        current_model->texcoord[i].s=(float)s/(float)model_hdr.skinWidth;
        current_model->texcoord[i].t=(float)t/(float)model_hdr.skinHeight;
    }

    fs_seek(fd,model_hdr.offsetTriangles,SEEK_SET);

    printf("Loading %i triangles\n",model_hdr.numTriangles);

    current_model->triangles=(struct triangle *)malloc(sizeof(struct triangle)*model_hdr.numTriangles);
    for(i=0;i<model_hdr.numTriangles;i++) {
        fs_read(fd,&current_model->triangles[i],sizeof(struct triangle));
    }


    /* read gl commands */
    fs_seek (fd, model_hdr.offsetGlCommands, SEEK_SET);

    printf("%i GL commands found...\n\n",model_hdr.numGlCommands);
    if (model_hdr.numGlCommands)
    {
        current_model->glCommandBuffer = (int *) malloc (sizeof (int) * model_hdr.numGlCommands);

        assert_msg(current_model->glCommandBuffer!=NULL,"md2 GL command OoM");

        fs_read (fd,current_model->glCommandBuffer, sizeof (int)* model_hdr.numGlCommands);
    }

    fs_close(fd);

    return (struct md2_model_node *)current_model;
}

#if 0
void draw_md2(struct md2_model_node *model, int texid, float mx, float my, float mz, int frame1, int frame2, int blendpos, int blendcount) {
    int i,j;

    unsigned short vert;
    unsigned short text;

    float x,y,z;
    float x1,y1,z1;
    float x2,y2,z2;




    switch_tex(texid);



    glBegin(GL_TRIANGLES);
    for(i=0;i<model->trianglecount;i++) {
        for(j=0;j<3;j++) {
            vert=model->triangles[i].vertex[j];
            text=model->triangles[i].texture[j];
            x1=model->frames[frame1].points[vert].x;
            x2=model->frames[frame2].points[vert].x;
            x=x1-((x1-x2)/blendcount)*blendpos;
            y1=model->frames[frame1].points[vert].y;
            y2=model->frames[frame2].points[vert].y;
            y=y1-((y1-y2)/blendcount)*blendpos;
            z1=model->frames[frame1].points[vert].z;
            z2=model->frames[frame2].points[vert].z;
            z=z1-((z1-z2)/blendcount)*blendpos;

            glTexCoord2f(model->texcoord[text].s,model->texcoord[text].t);

            kglVertex3f(x,y,z);
        }
    }
    glEnd();


}
#endif

/*
===============
R_LightPoint
===============
*/
void inline R_LightPoint (float ox,float oy, float oz, vec3_t color)
{

    color[0]=1.0f;

    color[1]=1.0f;
    color[2]=1.0f;



}

void draw_gl_md2(struct md2_model_node *model, int texid, float mx, float my, float mz, float angle_y, int frame1, int frame2, int blendpos, int blendcount) {

    unsigned short vert;

    float x,y,z;
    float x1,y1,z1;
    float x2,y2,z2;
    //    float l2,l1,l;

#ifdef MD2_LIGHT


#if 0

    shadedots = r_avertexnormal_dots[((int)(angle_y * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];

    an = angle_y/180*M_PI;
    shadevector[0] = fcos(-an);
    shadevector[1] = fsin(-an);
    shadevector[2] = 1.0f;

    //shadevector wird fr GL_DrawAliasShadow in Q2 benutzt!
    VectorNormalize (shadevector);
#endif
    shadelight[0]=0;
    shadelight[1]=0;
    shadelight[2]=0;

    R_LightPoint (my,my,mz, shadelight);
#endif

    switch_tex(texid);

    int i = 0;
    int val = model->glCommandBuffer[i++];

    while (val != 0)
    {
        int count;

        if (val > 0)
        {
            glBegin (GL_TRIANGLE_STRIP);
            count = val;
        }
        else
        {
            glBegin (GL_TRIANGLE_FAN);

            count = -val;
        }
        while (count--)
        {
            extern GLfloat vert_u,vert_v;//KGL-X gldraw.c

            vert_u = *(float *) &model->glCommandBuffer[i++];
            vert_v = *(float *) &model->glCommandBuffer[i++];


            vert = model->glCommandBuffer[i++];

            x1=model->frames[frame1].points[vert].x;
            x2=model->frames[frame2].points[vert].x;
            x=x1-((x1-x2)/blendcount)*blendpos;
            y1=model->frames[frame1].points[vert].y;
            y2=model->frames[frame2].points[vert].y;
            y=y1-((y1-y2)/blendcount)*blendpos;
            z1=model->frames[frame1].points[vert].z;
            z2=model->frames[frame2].points[vert].z;
            z=z1-((z1-z2)/blendcount)*blendpos;


#ifdef MD2_LIGHT
            extern int zclip;
            if(!zclip){

#if 0
                float nx1=model->frames[frame1].points[vert].nx;
                float nx2=model->frames[frame2].points[vert].nx;
                float nx=nx1-((nx1-nx2)/blendcount)*blendpos;

                nx1=model->frames[frame1].points[vert].ny;
                nx2=model->frames[frame2].points[vert].ny;
                float ny=nx1-((nx1-nx2)/blendcount)*blendpos;

                nx1=model->frames[frame1].points[vert].nz;
                nx2=model->frames[frame2].points[vert].nz;
                float nz=nx1-((nx1-nx2)/blendcount)*blendpos;



                kglColor4f (nx* shadelight[0], ny*shadelight[1], nz*shadelight[2],1.0f);

#endif


                l1=model->frames[frame1].points[vert].l;
                l2=model->frames[frame2].points[vert].l;
                l=(l1-((l1-l2)/blendcount)*blendpos);



                kglColor4f (l* shadelight[0], l*shadelight[1], l*shadelight[2],1.0f);
            }
#endif
            kglVertex3f(x,y,z);

        }

        glEnd ();

        val = model->glCommandBuffer[i++];
    }
}


//free the model memory
void free_md2(struct md2_model_node *model){


    DBG("freemd start\n");



    assert(model!=NULL);




    if (model->texcoord)
        free (model->texcoord);

    if (model->triangles)
        free (model->triangles);


    if (model->frames)
    {
        int i;

        for (i=0; i<model->framecount; i++)
            if (model->frames[i].points) free(model->frames[i].points);

        free(model->frames);
    }

    if (model->glCommandBuffer)
        free (model->glCommandBuffer);

    free(model);
    model=NULL;

    DBG("freemd end\n");
}




/*
 * center model
 *
 */
void
getBoundingBox_md2 (struct md2_model_node *model, float *minmax)
{int i;
	float minx, maxx;
	float miny, maxy;
	float minz, maxz;

	minx = miny = minz = 999999.0f;
	maxx = maxy = maxz = -999999.0f;

	/* get bounding box */
	for (i = 0; i < model->verticount; i++)
	{

        struct pt *v=&model->frames[0].points[i];

		if (v->x < minx)
			minx = v->x;
		else if (v->x > maxx)
			maxx = v->x;

		if (v->y < miny)
			miny = v->y;
		else if (v->y > maxy)
			maxy = v->y;

		if (v->z < minz)
			minz = v->z;
		else if (v->z > maxz)
			maxz = v->z;
	}

	minmax[0] = minx;
	minmax[1] = maxx;
	minmax[2] = miny;
	minmax[3] = maxy;
	minmax[4] = minz;
	minmax[5] = maxz;
}

float
getBoundingSphere_md2 (struct md2_model_node *model,
                       float *cx,float *cy,float *cz)
{

    float _cx=0,_cy=0,_cz=0;
    int i;

    for (i = 0; i < model->verticount; i++)
    {

        struct pt *pts=&model->frames[0].points[i];

        _cx+=pts->x;
        _cy+=pts->y;
        _cz+=pts->z;


    }
    _cx=_cx/model->verticount;
    _cy=_cy/model->verticount;
    _cz=_cz/model->verticount;


    float radius=0.0f,r2;

    if(cx!=NULL){
        *cx=_cx;
        *cy=_cy;
        *cz=_cz;
    }




    for (i = 0; i < model->verticount; i++){


        struct pt *pts=&model->frames[0].points[i];

        float x=pts->x-_cx;
        float y=pts->y-_cy;
        float z=pts->z-_cz;

        r2 = (x * x + y * y + z * z);
        if(r2 > radius)
            radius = r2;
    }



    return fsqrt(radius);
}

#if 0
void _makeFacetNormal (md2_model_t *model, md2_triangle_t *t, int frame, float *fn)
{
    float v1[3], v2[3];

    v1[0] = model->frames[frame].vertices[t->vertexIndices[1]].vertex[0] - model->frames[frame].vertices[t->vertexIndices[0]].vertex[0];
    v1[1] = model->frames[frame].vertices[t->vertexIndices[1]].vertex[1] - model->frames[frame].vertices[t->vertexIndices[0]].vertex[1];
    v1[2] = model->frames[frame].vertices[t->vertexIndices[1]].vertex[2] - model->frames[frame].vertices[t->vertexIndices[0]].vertex[2];

    v2[0] = model->frames[frame].vertices[t->vertexIndices[2]].vertex[0] - model->frames[frame].vertices[t->vertexIndices[0]].vertex[0];
    v2[1] = model->frames[frame].vertices[t->vertexIndices[2]].vertex[1] - model->frames[frame].vertices[t->vertexIndices[0]].vertex[1];
    v2[2] = model->frames[frame].vertices[t->vertexIndices[2]].vertex[2] - model->frames[frame].vertices[t->vertexIndices[0]].vertex[2];

    fn[0] = v1[1] * v2[2] - v1[2] * v2[1];
    fn[1] = v1[2] * v2[0] - v1[0] * v2[2];
    fn[2] = v1[0] * v2[1] - v1[1] * v2[0];
}



void _normalize (float *n)
{
    float l = (float) fsqrt (n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);

    if (l != 0.0f)
    {
        n[0] /= l;
        n[1] /= l;
        n[2] /= l;
    }
}


void md2_generateLightNormals (struct md2_model_node *model)
{
    int i;

    if (!model)
        return;


    for (i = 0; i < model->framecount; i++)
    {
        int j;

        /* clear all normals */
        for (j = 0; j < model->trianglecount; j++)
        {
            model->frames[i].vertices[j].normal[0] = 0.0f;
            model->frames[i].vertices[j].normal[1] = 0.0f;
            model->frames[i].vertices[j].normal[2] = 0.0f;
        }

        /* calc normals */
        for (j = 0; j < model->trianglecount; j++)
        {
            int k;
            float facetnormal[3];

            _makeFacetNormal (model, &model->triangles[j], i, facetnormal);

            for (k = 0; k < 3; k++)
            {
                model->frames[i].vertices[model->triangles[j].vertexIndices[k]].normal[0] -= facetnormal[0];
                model->frames[i].vertices[model->triangles[j].vertexIndices[k]].normal[1] += facetnormal[2];
                model->frames[i].vertices[model->triangles[j].vertexIndices[k]].normal[2] -= facetnormal[1];
            }
        }

        /* normalize normals */
        for (j = 0; j < model->header.numVertices; j++)
            _normalize (model->frames[i].vertices[j].normal);
    }
}
#endif
